// IN16-4-V1 voor PIC16F873A 
// Teruggebracht naar 4 buizen
// seconden resetten bij gelijkstellen

// V2.63
// Verbetering decimale punt

// V2.62
// Verbetering van DCF77 betrouwbaarheid

// V2.61
// bugfix DCF77 ging op output en werkt niet als ingang volgens gebruiker(s)

// V2.6 
// toevoegen van optie om 1 uur af te trekken van DCF77, i.v.m. UK locatie

// V2.5 
// onderdrukken van leading zero in 12hr formaat
// op nul stellen van timer variabelen, zodat de
// eerste seconde-tik pas valt *na* 1 seconde power on
// dit voorkomt verlopen bij aan/uitschakelen en RTC
// bruikbaar voor V1 hardware, zie #define

// V2.4 
// Nieuwe optie, 12/24 hr formaat 0=12 uur, 1=24uur

// V2.3 DS1203 RTC clock

// V2.2 poging om buizen te 'faden'

// save version clock3.c notsynced - knippert bij opstarten, totdat gelijk gezet wordt
// clock.c = werkfile, toevoegen van DCF decodering
// versie 1.2 -> clock 12.c div bugfix dcf77

//#define HARDWAREV1

#LIST
// Copyright(c) Frank Bemelman
// ---------------------------
// Auteur: Frank Bemelman
// Nixie Klok
// Software voor PIC16F873A 

// Kies hier uw processor
#define PROCESSOR16F873A

// ***** A T T E N T I E ******
//   char  -> 8 bit UNSIGNED !!!
//   int   -> 8 bit UNSIGNED !!!
//   long  -> 16 bit UNSIGNED !!!
//   short -> 1 bit !!!
//   signed int -> 8 bit signed !!!
//
//   De PIC is dus een hele rare omgeving !!!!
//
// DEFAULT data types zijn UNSIGNED !!!!!!!!!!!!!
// gebruik daarom (bijvoorbeeld) signed long & signed int


#define CRYSTAL 16000000

  // bij 16 mhz kristal, 200uS ms interrupt
  // reload 65536-800, 800 * 250nS, (250ns = 16mhz/4) 
  // 12uSec overhead, dus 48 minder 
  // maw 65536 - (800 - 48) = 64784
  // #define RELOADVALUE 64784
  // dit bleek +7 minuten op 660 minuten te snel, ofwel 1% te snel
  // nieuwe reload waarde wordt 65536 - (752 + 1%) -> 760

#define RELOADVALUE 64776

//#define RELOADVALUE 64500  /*60HZ*/

//#define RELOADVALUE 64300 /* 50HZ */

// Beschrijving van de aansluitingen op de PIC16C73A
// --------------------------------------------------------
//

// PIN 02 RA0 - Analoge ingang voor 0 - 5 volt DCLOW
// PIN 03 RA1 - sel3/ spkr
// PIN 04 RA2 - sel2
// PIN 05 RA3 - sw1
// PIN 06 RA4 - uitgang seldp
// PIN 07 RA5 - uitgang semicolon

#define TRISA 0b11001000 // '1' betekent input

#define SELDIG4      PIN_A0
#define SELDIG3      PIN_A1
#define SELDIG2      PIN_A2
#define SW1          PIN_A3
#define SELDP        PIN_A4
#define SELSEMICOLON PIN_A5


#define TRISC 0b00000000 // '1' betekent input
// PIN 11 RC0 - sel8
// PIN 12 RC1 - sel9
// PIN 13 RC2 - sel7
// PIN 14 RC3 - sel6
// PIN 15 RC4 - sel5
// PIN 16 RC5 - enable RS485 & spkr
// PIN 17 RC6 - tx
// PIN 18 RC7 - rx

#define DCF77    PIN_C7

#define RST1302   PIN_C5
#define SCLK1302  PIN_C6
#define IO1302    PIN_C7

#define SELDIG5  PIN_C4
#define SELDIG6  PIN_C3
#define SELDIG7  PIN_C2
#define SELDIG9  PIN_C1
#define SELDIG8  PIN_C0

// PIN 21 RB0 - sel0
// PIN 22 RB1 - sel1
// PIN 23 RB2 - seltube5
// PIN 24 RB3 - seltube6
// PIN 25 RB4 - seltube3
// PIN 26 RB5 - seltube4
// PIN 27 RB6 - seltube1
// PIN 28 RB7 - seltube2

#define TRISB 0b00001100 // '1' betekent input

#define SELTUB2  PIN_B7
#define SELTUB1  PIN_B6
#define SELTUB4  PIN_B5
#define SELTUB3  PIN_B4
#define ENCA     PIN_B3
#define ENCB     PIN_B2
#define SELDIG1  PIN_B1
#define SELDIG0  PIN_B0


// gebruikt om colonbits te setten
#define SELBYHEX0 0x40
#define SELBYHEX1 0x80
#define SELBYHEX2 0x10
#define SELBYHEX3 0x20
#define SELBYHEX4 0x04
#define SELBYHEX5 0x08


// flags als tijd veranderd is
#define CH_SECS1  0x01
#define CH_SECS10 0x02
#define CH_MINS1  0x04
#define CH_MINS10 0x08
#define CH_HRS1   0x10
#define CH_HRS10  0x20


// Include files en specifieke compiler #pragmas
// ---------------------------------------------

#include <c:\picc\devices\16F873A.H>

#FUSES HS,PROTECT,NOWDT,PUT,BROWNOUT, NOLVP
#USE FAST_IO(B)
#USE FAST_IO(C)

#use delay(clock=CRYSTAL)

unsigned int tclock[4];  // actuele tijd
unsigned int dclock[4];  // displaywaarde
unsigned int mclock[4];  // correctie
unsigned int colonbits;  // welke decimal points staan aan
unsigned int clockmode;  // wat staat er op de display

// clockmodes
#define CM_ALLOFF      0
#define CM_TIMERUN     1
#define CM_SETTIME     2
#define CM_SETCOR      3
#define CM_SETTUBESAVE 4
#define CM_SETOPTIONS  5

unsigned int keys[3]; // geeft tijd in ms dat toets is ingedrukt
int keyPressed[3];   // geeft aan dat toets ingedrukt is/was
#define KEYDOWN    3
#define KEYUP      2 
#define KEYSET     1  
#define KEYPOWER   0

// Globale declaraties
// -------------------

unsigned long StartUp;        // opstart vertraging (voor hardware teller)
unsigned long correction;     // correctie per uur, in msec (in stappen van ~9 sec per jaar)
unsigned long newcorrection;

unsigned int flasher;
unsigned int nextfase; // gaat elke msek
unsigned int fase;
unsigned int RunState; // geeft opstart, run, en uitschakel status aan.

unsigned int ClockAction;
#define WARMUP  0
#define POWERUP 1
#define RUNNING 2

// Eeprom stuff
// Eeprom (i2c) serial eeproms beslaan altijd 256 adressen
// om de tweede helft van de AT24C04 te adresseren
// moet A0 bit in device-adress byte aangepast worden
// wat we in dit programma niet nodig hebben en dus niet doen
unsigned int EepromAdr;		// adres binnen AT24C04 e2prom (1 kbit)
unsigned int EepromData;

#BYTE EEDATA = 0x10C
#BYTE EEADR  = 0x10D
#BYTE EEDATH = 0x10E
#BYTE EEADRH = 0x10F
#BYTE EECON1 = 0x18C
#BYTE EECON2 = 0x18D

#define OPTIONFADING    0x01
#define OPTIONFLASHNEON 0x02
#define OPTIONNEONON    0x04
#define OPTIONDCF77     0x11
#define OPTIONDCF77UK   0x12
#define OPTION1203      0x21
#define OPTION12HR      0x31


// Functie Prototypes
void do_keyboard(void); // elke msec
void HardInit(void);
unsigned int ReadFromE2prom(unsigned int adress);
void SaveToE2prom(unsigned int value, unsigned int adress);
void LoadParameters(void);
void SaveParms(void);
void settubeon(void);
void settubeoff(void);
void advance_clock(void);
void NewTimeForTubes(void);
void prepare_display(void);
void edit_time(void);
void setcor(void);
void settubesave(void);
void setoptions(void);
void flash_dp(void);
void do_dcf77(void);
void InitDS1203(void);
void GetTimeDS1203(void);
void SetTimeDS1203(void);
void WriteToDS1203(void);
void ReadFromDS1203(void);
void ConvertFromBCD(void);
void ConvertToBCD(void);
void ReadEncoder(void);
void ProcesEncoder(void);

// tijdens powerdown worden sommige waarden in e2prom snel opgeslagen
// hiervoor zijn onderstaande arrays in het leven geroepen.
// iPars wordt door het programma geactualiseerd bij nieuwe tellingen e.d.
// Bij powerdown, wordt Ipars vergeleken met ePars, en de verschillen
// worden dan opgeslagen.
// Parameters zoals ingesteld aantal worden steeds direct opgeslagen, het
// heeft geen zin om hiermee te wachten tot power down.

// specifieke eprom adressen:
#define AFIRST          0x20
#define ASETH           0x21  // correctie H
#define ASETL           0x22  // correctie L
#define ASTARTSAVEHR    0x23  // hoe laat powersave
#define ASTOPSAVEHR     0x24  // hoe laat powersave
#define AOPTIONS0       0x25  // options
#define AOPTIONSFF0     0x26  // options ^ 0xff
#define AOPTIONS1       0x27  // options
#define AOPTIONSFF1     0x28  // options ^ 0xff
#define AOPTIONS2       0x29  // options
#define AOPTIONSFF2     0x2A  // options ^ 0xff
#define AOPTIONS3       0x2B  // options
#define AOPTIONSFF3     0x2C  // options ^ 0xff
#define ALAST           0x2D  // altijd 1 hoger dan laatste adres definitie !!!!!!!!!!

unsigned int iPars[ALAST-AFIRST]; // initialiseren met e2prom vanaf 0x20 (ACNTL)
unsigned int ePars[ALAST-AFIRST]; // shadow van e2prom waarden

unsigned int tube;
unsigned int digit;
unsigned int prevdigit;
unsigned int seltube;
unsigned int second;

unsigned int chooseselect;
unsigned long key1;
unsigned int editdigit;
unsigned int keyreleased;
unsigned int flashdp;


unsigned int dutyselect;
signed long CompilerBugWorkaround; // deze variable gaat de mist in (page boundary ?)
unsigned int mSek1; 
unsigned int mSek10;
unsigned long uSec;
unsigned long prevkey;
unsigned int makecorrection;
unsigned int tempval;
unsigned int timechanged;
unsigned int minutes60;
unsigned int StopSaveHr;
unsigned int StartSaveHr;

unsigned int TimeSeconds;
unsigned int TimeMinutes;
unsigned int TimeHours;
unsigned int DimTubes;
unsigned int TubesOn;
unsigned int SpkrHigh;
unsigned int AlarmTone;
unsigned int NotSynced;
unsigned int Dcf77Pulse;
unsigned long Dcf77Pause;
unsigned int Dcf77Signal;
unsigned int Dcf77Bits;
unsigned int Dcf77Seconds;
unsigned int Dcf77Minutes;
unsigned int Dcf77Hours;
unsigned int Dcf77MinutesParity;
unsigned int Dcf77HoursParity;
unsigned long Dcf77PulsePause;
unsigned int Dcf77Filtered;
unsigned int Dcf77FilteredOld;

unsigned int FadeInLevel;
unsigned int FadeOutLevel;
unsigned int FadeSpeed;
unsigned int fclock[4];  // oude fadeout time
#define SPEEDOFFADE 40

unsigned int Read1302;
unsigned int Write1302;
unsigned int Command1302;
unsigned int ValTris;
unsigned int nop;
unsigned int Convert;

unsigned int NewEncoder;
unsigned int OldEncoder;
unsigned int ForwardCnt;
unsigned int ReverseCnt;


main()
{ 
  // initaliseer PIC hardware 
  HardInit(); // tris registers, schakeluitgangen, ad-conversie etc.

  // laad parameters als moduleadres, pwm-waarde, werkmodus etc.
  LoadParameters();

  if(Options & OPTION1203)  // rtc clock?
  { InitDS1203();
  }

  AlarmTone =1;
  
  second = 0;
  mSek10 = 0;
  uSec = 0;

  while (TRUE)
  { 

    if(second)
    { if(chooseselect!=CM_SETTIME)advance_clock();
      second = 0;
    }

    // eens per afgedraaid uur, correctiewaarde herinitialiseren
    if(timechanged & CH_MINS1)
    { timechanged &= CH_MINS1 ^ 0xff;
      minutes60++;
      if(minutes60==60)
      { minutes60 = 0;
        newcorrection += correction;
        newcorrection += 2031;
      }
      // benodigde correctie per (resterende) minuut bepalen
      tempval = (unsigned int)(newcorrection / (unsigned long)(60-minutes60));
      // max correctie is 9999, per minuut max 166
      makecorrection =  tempval;
      newcorrection -= tempval;

    }
    
    prepare_display();

    if(StartUp)
    { StartUp--;
      fase++;
      if(fase>3)fase=0;
      while(nextfase != fase);
    }
    else // teller is in werking
    { if(nextfase==fase) // tijd voor een nieuwe actie, elke msek
      { switch(fase)
        { case 0:      // update van rode display

            if(DimTubes) // nachtstand, display uit
            { if(key1)   // toets?
              { // forceer display aan
                TubesOn =  100; // 50 x 10ms display
              }
              else if (TubesOn<50)
              { TubesOn =  0; // display uit
              }
            }
            else
            { TubesOn =  50; // 50 x 10ms display
            }


            break;
          case 1:     // update van groene display
            break;
          case 2:     // teller == preset ????
            Flasher++;
            if(Flasher>125)Flasher=0; // halve seconde
            break;

          case 3: // elke 4 msek deze state uitvoeren    
            ReadEncoder();
            if(OldEncoder != NewEncoder)
            { ProcesEncoder();
              OldEncoder = NewEncoder;
            } 
            break;

          default:
            break;
        }

        switch(ClockAction)
        { case POWERUP:
            clockmode = CM_TIMERUN;
            ClockAction = RUNNING;
            break;
          default:
            break;
        }
        do_keyboard();
        do_dcf77();
        fase++;
        if(fase>3)fase=0;
        if(FadeSpeed)FadeSpeed--;
        else
        { FadeSpeed = SPEEDOFFADE;
          if(FadeInLevel)
          { FadeInLevel--;
            FadeOutLevel++;
          }
        }
      }
    }
  }
}  


void do_keyboard(void) // elke msec
{ if(INPUT(SW1)==0)
  { if(key1<5000)key1++;
  }
  else
  { if(chooseselect == CM_ALLOFF)
    { if(clockmode != CM_TIMERUN)
      { chooseselect = clockmode;
      }
    }
    prevkey = key1;
    key1 = 0;
    keyreleased = 1;
  }

  // druk knop in, na 5 seconde, eerste keuze, na elke volgende 2 seconde, volgende keuze
  if(chooseselect == CM_ALLOFF)
  { if(key1==100)
    { NotSynced = 0;
    }
    if(clockmode == CM_TIMERUN)
    { if(key1==5000)
      { clockmode = CM_SETTIME;
        NotSynced = 0;
        key1=1;
        colonbits = SELBYHEX0;
        editdigit = 0;
      }
    }
    else if(clockmode == CM_SETTIME)
    { if(key1==2000)
      { clockmode = CM_SETCOR;
        key1=1;
        colonbits = SELBYHEX1;
        editdigit = 2;
      }
    }
    else if(clockmode == CM_SETCOR)
    { if(key1==2000)
      { clockmode = CM_SETTUBESAVE;
        key1=1;
        colonbits = SELBYHEX2;
        editdigit = 1;
      }
    }
    else if(clockmode == CM_SETTUBESAVE)
    { if(key1==2000)
      { clockmode = CM_SETOPTIONS;
        key1=1;
        colonbits = SELBYHEX3;
        editdigit = 0;
      }
    }
  }
  else // choose select bevat keuze
  { flash_dp();
    if(prevkey>25)
    { if(prevkey<250) // korte toets
      { if(chooseselect==CM_SETTIME)edit_time();
        else if(chooseselect==CM_SETCOR)setcor();
        else if(chooseselect==CM_SETTUBESAVE)settubesave();
        else if(chooseselect==CM_SETOPTIONS)setoptions();
      }
    }    
    else
    { if(key1==1000)
      { editdigit++;
        if(editdigit>3)
        { editdigit=0;
        }
      }
      if(key1==5000)
      { if(chooseselect==CM_SETCOR)SaveParms();
        if(chooseselect==CM_SETTIME)
        { mSek10 = 0;
          Second = 0;
          TimeSeconds = 0;
          TimeMinutes = tclock[2] * 10;
          TimeMinutes += tclock[3];
          TimeHours   = tclock[0] * 10;
          TimeHours   += tclock[1];
          advance_clock();
          advance_clock();
          advance_clock();
          advance_clock();
          advance_clock();
          if(Options & OPTION1203)  // rtc clock?
          { SetTimeDS1203();
          }
        }
        if(chooseselect==CM_SETTUBESAVE)SaveParms();
        if(chooseselect==CM_SETOPTIONS)SaveParms();
        clockmode = CM_TIMERUN;
        chooseselect = CM_ALLOFF;
        flashdp = 0;
        key1=1;
      }
    }

  }

}

#INT_TIMER1
timer1_handler(void)
{ 
  SET_TIMER1(RELOADVALUE); 

  dutyselect++;
  if(dutyselect>0x09)
  { // volgende buis
    dutyselect=0;
    tube>>=1;
    if(tube<0x10)
    { tube = 0x80;  // begin bij RB7, tube2, 2de buis van links, enkele uren
      seltube = 1;
    }
    else
    { if(seltube==1)seltube=0;
      else if (seltube==0)seltube=3;
      else if (seltube==3)seltube=2;
    }

    if(seltube==0)
    { if(mSek10<15)
      { if(chooseselect == CM_ALLOFF) // we zitten niet in een menu
        { if(Options & OPTIONCOLONS)  // flashing semicolon neons wanted
          { output_high(SELSEMICOLON);
          }
        }
      }
      else
      { output_low(SELSEMICOLON);
      }
    }
    else
    { output_low(SELSEMICOLON);
    }
  }


  // doe de tubes
  if(dutyselect==0)
  { settubeoff(); // cathodes allemaal floating
  }

  if(dutyselect>0x07)
  { output_low(SELTUB1);
    output_low(SELTUB2);
    output_low(SELTUB3);
    output_low(SELTUB4);
    output_high(SELDP);
    output_high(SELDIG0);
    output_high(SELDIG1);
    output_high(SELDIG2);
    output_high(SELDIG3);
    output_high(SELDIG4);
    output_high(SELDIG5);
    output_high(SELDIG6);
    output_high(SELDIG7);
    output_high(SELDIG8);
    output_high(SELDIG9);
  }
  else // dutycycle 0 t/m 7
  { if(TubesOn)
    { // turn on anode
      if(seltube == 0)output_high(SELTUB1);
      if(seltube == 1)output_high(SELTUB2);
      if(seltube == 2)output_high(SELTUB3);
      if(seltube == 3)output_high(SELTUB4);

      // is dit de buis die een DP moet tonen? 
      if(colonbits & tube)
      { // dp inverse knipperen met digit
        // zet de dp aan
        //if(flashdp==0)output_high(SELDP);
        //else
        { if(Flasher>100)
          { settubeoff();            // cijfer van buis uit
            output_high(SELDP);      // dp aan
          }
          else
          { digit = dclock[seltube]; // dit is het cijfer van de nieuwe buis
            settubeon(); // cathodes van buis 'Digit' on
          }
        }
      }
      else // normaal bedrijf
      { if(dutyselect>=FadeInLevel) // de nieuwe tijd displayen
        { digit = dclock[seltube]; // dit is het cijfer van de nieuwe buis
          settubeon(); // cathodes van buis 'Digit' on
        }

        else //if(dutyselect>=FadeOutLevel) // de oude tijd displayen
        { digit = fclock[seltube]; // dit is het cijfer van de oude buis
          settubeon(); // cathodes van buis 'Digit' on
        }
      }
    }
  }
}




unsigned int ReadFromE2prom(unsigned int adress)
{ unsigned int data;

  EEADR = adress;
  BIT_CLEAR(EECON1, 7); // EEPGD, R/W
  BIT_SET(EECON1, 0); // EEPGD, RD
  data = EEDATA;

  return data;
}

void SaveToE2prom(unsigned int value, unsigned int adress)
{ 
  EEADR = adress;
  EEDATA = value;
  BIT_CLEAR(EECON1, 7); // EEPGD, R/W
  BIT_SET(EECON1, 2); // EEPGD, WREN
  DISABLE_INTERRUPTS(GLOBAL);
  EECON2 = 0x55;
  EECON2 = 0xAA;
  BIT_SET(EECON1, 1); // EEPGD, WR
  ENABLE_INTERRUPTS(GLOBAL);
  DELAY_MS(15);             // vlgs specs 4 ms typ, 8 ms max, 15 mS moet zeker voldoende zijn
}




void HardInit(void)
{ int n;
  // initialiseer port direction registers
  // zie #defines in I/O beschrijving
  SET_TRIS_A(TRISA); 

  SET_TRIS_B(TRISB); 
  port_b_pullups(TRUE);

  ValTris = 0x80;
  SET_TRIS_C(ValTris); 

  output_low(RST1302);

  // alle segmenten uit
  output_low(SELDIG2);
  output_low(SELDIG3);
  output_low(SELDIG4);
  output_low(SELDIG5);
  output_low(SELDIG6);
  output_low(SELDIG7);
  output_low(SELDIG8);
  output_low(SELDIG9);
  output_low(SELSEMICOLON);
  output_low(SELDP);

  ClockAction = POWERUP;
  clockmode = CM_ALLOFF;
  StartUp=1000;                         // opstart/inregel vertraging
  makecorrection = 0;
  chooseselect = CM_ALLOFF;
  flashdp = 0;
  timechanged = 0;
  minutes60 = 59;
  newcorrection = 0;

  TimeSeconds = 0;
  TimeMinutes = 0;
  TimeHours   = 12;

  NotSynced = 1;

  for(n=0;n<6;n++)
  { fclock[n] = 12;
    tclock[n] = 12;
  }

  // initaliseer TIMER1 voor display multiplex via interrupt
  SETUP_TIMER_1(T1_INTERNAL| T1_DIV_BY_1);
  SET_TIMER1(RELOADVALUE);
  ENABLE_INTERRUPTS(INT_TIMER1);

  // 16000000/4->4000000 / 8 -> 2uS
  // int = 256 * 2 uS -> 512uS
  setup_counters(RTCC_INTERNAL, RTCC_DIV_8);
  ENABLE_INTERRUPTS(INT_RTCC);

  ENABLE_INTERRUPTS(GLOBAL);
  Dcf77Bits = 60;
  Dcf77Minutes = 99;
  Dcf77Hours = 99;
  Dcf77MinutesParity = 1;
  Dcf77HoursParity = 1;

  ReadEncoder();
  OldEncoder = NewEncoder;
  ForwardCnt = 0;
  ReverseCnt = 0;

}


void SaveParms(void)
{ unsigned int n; 

  iPars[ASETH-AFIRST] = Correction / 256;
  iPars[ASETL-AFIRST] = Correction % 256;
  iPars[ASTARTSAVEHR-AFIRST] = StartSaveHr;
  iPars[ASTOPSAVEHR-AFIRST]  = StopSaveHr;
  iPars[AOPTIONS-AFIRST]  = Options;
  iPars[AOPTIONSFF-AFIRST]  = Options ^ 0xff;

  for(n=AFIRST; n<ALAST; n++)
  { if(iPars[n-AFIRST] != ePars[n-AFIRST])
    { SaveToE2prom(iPars[n-AFIRST], n);
      ePars[n-AFIRST] = iPars[n-AFIRST];
    }
  }

}

void LoadParameters(void)
{ unsigned int n;

  // lees opgeslagen waarden
  for(n=AFIRST; n<ALAST; n++)
  { iPars[n-AFIRST] = ReadFromE2prom(n);
    ePars[n-AFIRST] = ReadFromE2prom(n);
  }

  Correction  = iPars[ASETH-AFIRST] * 256;
  Correction += iPars[ASETL-AFIRST];
  if(Correction>9999)Correction = 5000;

  mclock[0]=(unsigned int)(correction/1000L);
  mclock[1]=(unsigned int)((correction/100L)%10L);
  mclock[2]=(unsigned int)((correction/10L)%10L);
  mclock[3]=(unsigned int)(correction%10L);


  StartSaveHr = iPars[ASTARTSAVEHR-AFIRST];
  StopSaveHr =  iPars[ASTOPSAVEHR-AFIRST];
  if(StartSaveHr>23)
  { StartSaveHr=0;
    StopSaveHr=0;
  }
  if(StopSaveHr>23)
  { StartSaveHr=0;
    StopSaveHr=0;
  }

  Options = iPars[AOPTIONS-AFIRST];
  if(Options != (iPars[AOPTIONSFF-AFIRST] ^ 0xff))
  { Options = OPTIONFADING | OPTIONCOLONS | OPTION1203;
    SaveParms();
  }

  // neem altijd aan dat ds1203 er is
  Options |= OPTION1203;

}

void settubeon(void)
{ if(Digit==0)        output_high(SELDIG0);
  else if (Digit==1)  output_high(SELDIG1);
  else if (Digit==2)  output_high(SELDIG2);
  else if (Digit==3)  output_high(SELDIG3);
  else if (Digit==4)  output_high(SELDIG4);
  else if (Digit==5)  output_high(SELDIG5);
  else if (Digit==6)  output_high(SELDIG6);
  else if (Digit==7)  output_high(SELDIG7);
  else if (Digit==8)  output_high(SELDIG8);
  else if (Digit==9)  output_high(SELDIG9);
}


void settubeoff(void)
{ output_low(SELDIG0);
  output_low(SELDIG1);
  output_low(SELDIG2);
  output_low(SELDIG3);
  output_low(SELDIG4);
  output_low(SELDIG5);
  output_low(SELDIG6);
  output_low(SELDIG7);
  output_low(SELDIG8);
  output_low(SELDIG9);
  output_low(SELDP);
}


void advance_clock(void)
{ TimeSeconds++;
  timechanged |= CH_SECS1;
  if(TimeSeconds>59)
  { TimeSeconds = 0;
    TimeMinutes++;
    timechanged |= CH_MINS1;
    if(TimeMinutes>59)
    { TimeMinutes = 0;
      TimeHours++;
      timechanged |= CH_HRS1;
      if(TimeHours>23)
      { TimeHours = 0;
      }
      if(Options & OPTION1203)  // rtc clock?
      { SetTimeDS1203();
      }
    }
  }

  NewTimeForTubes();
}

void NewTimeForTubes(void)
{ unsigned int hrs;
  
  fclock[0] = tclock[0];
  fclock[1] = tclock[1];
  fclock[2] = tclock[2];
  fclock[3] = tclock[3];

  hrs = TimeHours;
  //if(Options & OPTION12HR)  // 12 hour format clock?
  //{ hrs = TimeHours;
  //  if(hrs==0)hrs+=12;
  //  if(hrs>12)
  //  { hrs -= 12;
  //  }
  //}

  tclock[3] = TimeMinutes % 10;
  tclock[2] = TimeMinutes / 10;
  tclock[1] = hrs % 10;
  tclock[0] = hrs / 10;

  //if(Options & OPTION12HR)  // 12 hour format clock?
  //{ if(tclock[0]==0)tclock[0]=12; // display geen leading zero
  //}
      

  // nieuw cijfer, begint op lage sterkte
  FadeSpeed = SPEEDOFFADE;

  if(clockmode == CM_TIMERUN)
  { if(Options & OPTIONFADING)
    { if(NotSynced == 0)
      { FadeInLevel = 7;
        FadeOutLevel = 0;
      }
    }
  }

}

void prepare_display(void)
{ unsigned int n; 
 
  // in principe buizen aan
  DimTubes = 0; 

  switch(clockmode)
  { case CM_TIMERUN:
      dclock[0]=tclock[0];
      dclock[1]=tclock[1];
      dclock[2]=tclock[2];
      dclock[3]=tclock[3];
      colonbits = 0;

      // tenzij...
      if(StartSaveHr>StopSaveHr)
      { // 21 - 19
        if(TimeHours>=StartSaveHr)DimTubes =  1;
        if(TimeHours<StopSaveHr)DimTubes =  1;
      }
      
      if(StartSaveHr<StopSaveHr)
      { // 19 - 21
        if(TimeHours>=StartSaveHr)
        { if(TimeHours<StopSaveHr)DimTubes =  1;
        }
      }

      // tenzij...
      if(NotSynced) 
      { if(mSek10>50) // laten knipperen
        { DimTubes =  1; 
        }
      }
      break;

    case CM_SETTIME:
      dclock[0]=tclock[0];
      dclock[1]=tclock[1];
      dclock[2]=tclock[2];
      dclock[3]=tclock[3];
      break;

    case CM_SETCOR:
      dclock[0]=mclock[0];
      dclock[1]=mclock[1];
      dclock[2]=mclock[2];
      dclock[3]=mclock[3];
      break;

    case CM_SETTUBESAVE:
      dclock[0]=StartSaveHr/10;
      dclock[1]=StartSaveHr%10;
      dclock[2]=StopSaveHr/10;
      dclock[3]=StopSaveHr%10;
      break;

    case CM_SETOPTIONS:
      for(n=0;n<4;n++)
      { if(Options & (1<<n))
        { dclock[n] = 1;
        }
        else
        { dclock[n] = 0;
        }
      }
      break;


    default:
      // display niets
      dclock[0]=12;
      dclock[1]=12;
      dclock[2]=12;
      dclock[3]=12;
      colonbits = SELBYHEX0;
      break;
  }  

}

void edit_time(void)
{ switch(editdigit)
  { case 0:
      if(tclock[0]==2)
      { if(tclock[1]>3)tclock[1]=3;     
        tclock[0]=255;
      }     
      tclock[0]++;     
      break;
    case 1:
      if(tclock[0]==2)
      { if(tclock[editdigit]==3)
        { tclock[editdigit]=9;
        }
      }
      if(tclock[editdigit]==9)
      { tclock[editdigit]=255;     
      }     
      tclock[editdigit]++;     
      break;
    case 3:
      if(tclock[editdigit]==9)
      { tclock[editdigit]=255;     
      }     
      tclock[editdigit]++;     
      break;
    case 2:
      if(tclock[editdigit]==5)
      { tclock[editdigit]=255;     
      }     
      tclock[editdigit]++;     
      break;
  }
}

void settubesave(void)
{ 
  switch(editdigit)
  { case 0:
    case 1:
      StartSaveHr++;
      if(StartSaveHr>23)StartSaveHr=0;
      break;

    case 2:
    case 3:
      StopSaveHr++;
      if(StopSaveHr>23)StopSaveHr=0;
      break;
  }

}

void setcor(void)
{ 
  if(mclock[editdigit]>=9)
  { mclock[editdigit]=255;     
  }     
  mclock[editdigit]++;

  correction = (unsigned long)mclock[0] * 1000L;
  correction += (unsigned long)mclock[1] * 100L;
  correction += (unsigned long)mclock[2] * 10L;
  correction += (unsigned long)mclock[3];

}

void setoptions(void)
{ unsigned int n; 
  n = 1 << editdigit;
  if(Options & n)
  { Options &= n ^ 0xff;
  }
  else
  { Options |= n;
    if(n==OPTION1203)InitDS1203();
  }
}


void flash_dp(void)
{ flashdp = 2;
  switch(editdigit)
  { case 0:
      colonbits = SELBYHEX0;
      break;
    case 1:
      colonbits = SELBYHEX1;
      break;
    case 2:
      colonbits = SELBYHEX2;
      break;
    case 3:
      colonbits = SELBYHEX3;
      break;
  }
}


#INT_RTCC
timer2_handler(void)
{ // Deze interrupt komt elke 0,5 milliseconde. 

  // door 511 ipv 512 loopt de klok per uur 7,031 seconde te langzaam
  // want (511/512) * 3600 maakt dat tekort
  uSec+=511;

  if(!TubesOn)
  { if(SpkrHigh)
    { SpkrHigh = 0;
    }
    else
    { SpkrHigh = 1;
    }
  }

  if(uSec>1000)
  { uSec -=1000;
    mSek1++;
    if(mSek1>0x09)
    { mSek1=0;
      mSek10++;

      if(TubesOn)TubesOn--;

      // nog corrigeren per minuut, max 166, max 3mS per seconde
      if(makecorrection)
      { makecorrection--;
        uSec+=1000;
      }

      if(mSek10>99)
      { mSek10=0;
        second=1;
      }
    }
    nextfase++;
    if(nextfase>3)nextfase=0;
  }
}

void do_dcf77(void)
{ // pulsen zijn negatief, en de neergaande flank is het begin van een seconde
  // de opgaande flank geeft de lengte aan, 100ms of 200ms. Een pauze van 2
  // seconde is begin van de minuut

//  Dcf77Bits = TimeSeconds;

  if(INPUT(DCF77)==1)
  { if(!Dcf77Filtered) // signaal is laag
    { if(Dcf77Signal<25)Dcf77Signal++;
      else
      { Dcf77Filtered = 1;
      }
    }
    else
    { Dcf77Signal = 25;
    }
  }
  else
  { if(Dcf77Filtered) // signaal is hoog
    { if(Dcf77Signal)Dcf77Signal--;
      else
      { Dcf77Filtered = 0;
      }
    }
    else
    { Dcf77Signal = 0;
    }
  }

  if(Dcf77Filtered)
  { // tijdmeting pulse
    Dcf77Pause++;
  }
  else Dcf77Pulse++;

  if(Dcf77Filtered != Dcf77FilteredOld)
  { Dcf77FilteredOld = Dcf77Filtered;

    if(Dcf77Filtered)
    { // opgaande flank
      Dcf77Pause = 0;

      if(Dcf77Bits==20)
      { Dcf77MinutesParity = 0;
        Dcf77HoursParity = 0;
        Dcf77Minutes = 0;
        Dcf77Hours = 0;
      }

      if((Dcf77Pulse>150) && (Dcf77Pulse<250))
      { //output_high(PIN_A1);
        if(Dcf77Bits==21)
        { Dcf77Minutes = 1; 
          Dcf77MinutesParity++;
        }   
        else if(Dcf77Bits==22)
        { Dcf77Minutes += 2; 
          Dcf77MinutesParity++;
        }   
        else if(Dcf77Bits==23)
        { Dcf77Minutes += 4; 
          Dcf77MinutesParity++;
        }   
        else if(Dcf77Bits==24)
        { Dcf77Minutes += 8; 
          Dcf77MinutesParity++;
        }   
        else if(Dcf77Bits==25)
        { Dcf77Minutes += 10; 
          Dcf77MinutesParity++;
        }   
        else if(Dcf77Bits==26)
        { Dcf77Minutes += 20; 
          Dcf77MinutesParity++;
        }   
        else if(Dcf77Bits==27)
        { Dcf77Minutes += 40; 
          Dcf77MinutesParity++;
        }   
        else if(Dcf77Bits==28)
        { Dcf77MinutesParity++;
        }   

        if(Dcf77Bits==29)
        { Dcf77Hours = 1; 
          Dcf77HoursParity++;
        }   
        else if(Dcf77Bits==30)
        { Dcf77Hours += 2; 
          Dcf77HoursParity++;
        }   
        else if(Dcf77Bits==31)
        { Dcf77Hours += 4; 
          Dcf77HoursParity++;
        }   
        else if(Dcf77Bits==32)
        { Dcf77Hours += 8; 
          Dcf77HoursParity++;
        }   
        else if(Dcf77Bits==33)
        { Dcf77Hours += 10; 
          Dcf77HoursParity++;
        }   
        else if(Dcf77Bits==34)
        { Dcf77Hours += 20; 
          Dcf77HoursParity++;
        }   
        else if(Dcf77Bits==35)
        { Dcf77HoursParity++;
        }   
      }
      else
      { 
      }
      Dcf77Bits++; 
    }
    else
    { // neergaande flank
      // twee pulsen bij elkaar optellen, maar de korte puls kan kort (100ms) of lang (200ms) zijn
      Dcf77PulsePause = Dcf77Pulse + Dcf77Pause;
      Dcf77Pulse = 0;

      if((Dcf77PulsePause>1900) && (Dcf77PulsePause<2100))
      { Dcf77Bits = 0;
        Dcf77Seconds = 0;
//        mSek10 = 0;
//        Second = 0;
//        TimeSeconds=0;
//        tclock[5] = TimeSeconds % 10;
//        tclock[4] = TimeSeconds / 10;

        if(Dcf77Minutes<60) // sanity check 1
        { if(Dcf77Hours<24) // sanity check 2
          { if(Dcf77Minutes>0) // sanity check 3
            { if(Dcf77Hours>0) // sanity check 4
              { if(!(Dcf77MinutesParity & 0x01)) // sanity check 3
                { if(!(Dcf77HoursParity & 0x01)) // sanity check 4
                  { TimeMinutes = Dcf77Minutes;
                    tclock[3] = TimeMinutes % 10;
                    tclock[2] = TimeMinutes / 10;
                    if(Options & OPTIONDCF77UK)
                    { if(Dcf77Hours == 0)Dcf77Hours = 23;
                      else Dcf77Hours--;
                    }
                    TimeHours = Dcf77Hours;
                    tclock[1] = TimeHours   % 10;
                    tclock[0] = TimeHours   / 10;

        mSek10 = 0;
        Second = 0;
        TimeSeconds=0;

                    NotSynced = 0;
                  }
                }
              }
            }
          }
        }
        Dcf77Minutes = 99;
        Dcf77Hours = 99;
        Dcf77MinutesParity = 1;
        Dcf77HoursParity = 1;
      }
//      else if((Dcf77PulsePause>950) && (Dcf77PulsePause<1050))
//      { // keurige puls
//      }     
    }
  } 
}

void InitDS1203(void)
{  

  Command1302 = 0x8e; // controlregister, write
  Write1302 = 0x0;  // write protect off
  WriteToDS1203();

  // test of er een DS1203 aangesloten is
  Command1302 = 0xc0; // ram adres 0, write
  Write1302 = 0xAA;   // pattern
  WriteToDS1203();
  ReadFromDS1203();
  if(Read1302 != 0xAA)
  { // geen 1302 aanwezig, uit de config halen
    Options &= OPTION1203 ^ 0xff;
    SaveParms();
    return;
  }


  Command1302 = 0x84; // hours
  ReadFromDS1203();
  if(Read1302 & 0x80) // clock staat in 12hr mode!
  { Write1302 = Read1302; // uren overnemen
    Write1302 &= 0x7f; // 12hr mode uitschakelen
    WriteToDS1203();
  }

  // read seconds
  Command1302 = 0x80; // seconds
  ReadFromDS1203();
  if(Read1302 & 0x80) // clock staat stil!
  { Write1302 = Read1302; // seconds overnemen
    Write1302 &= 0x7f;
    WriteToDS1203();
  }
  else
  { GetTimeDS1203();
    NotSynced = 0;
  }

  Command1302 = 0x90; // trickle charger
  Write1302 = 0xA5; // 1 diode charger, 2K
  WriteToDS1203();

}

void GetTimeDS1203(void)
{ Command1302 = 0x80; // seconds
  ReadFromDS1203();
  ConvertFromBCD();
  TimeSeconds = Convert;  

  Command1302 = 0x82; // minutes
  ReadFromDS1203();
  ConvertFromBCD();
  TimeMinutes = Convert;

  Command1302 = 0x84; // hours
  Read1302 &= 0x7f;   // mask 12/24 mode bit
  ReadFromDS1203();
  ConvertFromBCD();
  TimeHours = Convert;
}

void SetTimeDS1203(void)
{ Command1302 = 0x80; // seconds
  Convert = TimeSeconds;    
  ConvertToBCD();
  WriteToDS1203();

  Command1302 = 0x82; // minutes
  Convert = TimeMinutes;    
  ConvertToBCD();
  WriteToDS1203();

  Command1302 = 0x84; // hours
  Convert = TimeHours;    
  ConvertToBCD();
  WriteToDS1203();
}


void WriteToDS1203(void)
{ unsigned int n;

#ifdef HARDWAREV1
  ValTris = TRISA;
  SET_TRIS_A(ValTris); 
#else
  ValTris = 0x0;
  SET_TRIS_C(ValTris); 
#endif

    nop++;
    nop++;

  output_low(SCLK1302);
    nop++;
    nop++;
  output_high(RST1302);
    nop++;
    nop++;

  Command1302 &= 0xfe; // maak er een write van

  n = 0x01;
  while(n)
  { if(Command1302 & n)output_high(IO1302);
    else             output_low(IO1302);
    output_high(SCLK1302);
    nop++;
    nop++;
    output_low(SCLK1302);
    nop++;
    nop++;
    n <<= 1;
  }

  n = 0x01;
  while(n)
  { if(Write1302 & n)output_high(IO1302);
    else             output_low(IO1302);
    output_high(SCLK1302);
    nop++;
    nop++;
    output_low(SCLK1302);
    nop++;
    nop++;
    n <<= 1;
  }

  output_low(RST1302);

}

void ReadFromDS1203(void)
{ unsigned int n;

#ifdef HARDWAREV1
  ValTris = TRISA;
  SET_TRIS_A(ValTris); 
#else
  ValTris = 0x0;
  SET_TRIS_C(ValTris); 
#endif

    nop++;
    nop++;

  output_low(SCLK1302);
    nop++;
    nop++;
  output_high(RST1302);
    nop++;
    nop++;

  Command1302 |= 0x01; // maak er een read van

  n = 0x01;
  while(n)
  { if(Command1302 & n)output_high(IO1302);
    else             output_low(IO1302);
    output_high(SCLK1302);
    nop++;
    nop++;
    output_low(SCLK1302);
    nop++;
    nop++;
    n <<= 1;
  }

#ifdef HARDWAREV1
  ValTris = TRISA | 0x04;
  SET_TRIS_A(ValTris); 
#else
  ValTris = 0x80;
  SET_TRIS_C(ValTris); 
#endif
    nop++;
    nop++;

  n = 0x01;
  while(n)
  { if(INPUT(IO1302)==1)Read1302 |= n;
    else                Read1302 &= (n ^ 0xff);
    output_high(SCLK1302);
    nop++;
    nop++;
    output_low(SCLK1302);
    nop++;
    nop++;
    n <<= 1;
  }

  output_low(RST1302);

}

void ConvertFromBCD(void)
{ Convert = Read1302;
  Read1302 &= 0xf0;
  while(Read1302)
  { Convert -= 6;
    Read1302 -= 16;
  }
}

void ConvertToBCD(void)
{ Write1302 = Convert/10;
  Write1302 <<= 4;
  Write1302 += Convert%10;
}

void ReadEncoder(void)
{ 
  // is geen atomic read, maar ja...
  NewEncoder = INPUT(ENCA);
  if(INPUT(ENCB)==1)NewEncoder |= 0x02;
  else              NewEncoder &= 0x02 ^ 0xff; 

}

void ProcesEncoder(void)
{ if((NewEncoder==0x00) && (OldEncoder==0x01))ForwardCnt++;
  if((NewEncoder==0x01) && (OldEncoder==0x03))ForwardCnt++;
  if((NewEncoder==0x03) && (OldEncoder==0x02))ForwardCnt++;
  if((NewEncoder==0x02) && (OldEncoder==0x00))ForwardCnt++;

  if((NewEncoder==0x00) && (OldEncoder==0x02))ReverseCnt++;
  if((NewEncoder==0x01) && (OldEncoder==0x00))ReverseCnt++;
  if((NewEncoder==0x03) && (OldEncoder==0x01))ReverseCnt++;
  if((NewEncoder==0x02) && (OldEncoder==0x03))ReverseCnt++;

  if(ForwardCnt>3)
  { ForwardCnt = 0;
    TimeSeconds = 0;
    TimeMinutes++;
    if(TimeMinutes>59)
    { TimeMinutes = 0;
      TimeHours++;
      if(TimeHours>23)
      { TimeHours = 0;
      }
    }

    NewTimeForTubes();

    if(Options & OPTION1203)  // rtc clock?
    { SetTimeDS1203();
    }
  }

  if(ReverseCnt>3)
  { ReverseCnt = 0;
    TimeSeconds = 0;
    if(TimeMinutes==0)
    { TimeMinutes = 59;
      if(TimeHours == 0)
      { TimeHours = 23;
      }
      else 
      { TimeHours--;
      }
    }
    else
    { TimeMinutes--;
    }      

    NewTimeForTubes();

    if(Options & OPTION1203)  // rtc clock?
    { SetTimeDS1203();
    }
  }
}

// End of file //

